-- Terrain Generator for Civ VI Got Lakes Map Script
-- Author: Scrum Lord
------------------------------------------------------------------------------
include "GotLakes_Options"
include "GotLakes_Utils"

GotLakes_TerrainGenerator = {};

function GotLakes_TerrainGenerator.Create(
	climateArgs,
	biomeArgs,
	sunlightArgs,
	donutFillingMap,
	createClimateFracFcn,
	latitudeFcn)
	-- Terrain-Gen constructor.
	-- Args:
	--   climateArgs: info for chosen climate
	--   biomeArgs: info for chosen biome
	--   sunlightArgs: info for chosen sunlight
	--   donutFillingMap: map indicating which plots are donut filling
	--   createClimateFracFcn: function to create a climate fractal given the fractal's name
	--   latitudeFcn: function to get latitude at (x, y)
	local iW, iH = GetMapGridSize();
	
	local temperature = UserInput.temperature;
	local tundraLevel = UserInput.tundraLevel;
	local grassPlains = UserInput.grassPlains;
	local climate     = UserInput.climate;
	local sunlight    = UserInput.sunlight;
	local biome       = UserInput.biome;
	
	-- Special fractal for the poles of a regional tilted axis.
	-- Because solid tundra/snow/sand is not a fun place to start a game.
	local coldPolePct = sunlightArgs.coldPolePct[climate];
		
	-- Deserts are processed first, so Plains will take this 
	-- percentage of whatever remains. - Bob
	local desertBasePct = sunlightArgs.desertBasePct[climate];
	local desertPctOffset = 0;
	local desertPctShift = 0;
	
	if (sunlightArgs.enableDesertPctShift) then
		-- Enable desert fractal adjustments for certain types of sunlight.
		desertPctOffset = biomeArgs.desertPctOffset;
		desertPctShift  = 10;
	end
	local desertPct = desertBasePct + desertPctOffset;
	
	-- Plains vs Grassland.
	local plainsPct = biomeArgs.plainsPct;
	
	-- How much to increase the world's min/max latitude based on temperature.
	local temperatureLatShift = 0.1;
	
	-- How much to adjust various lat thresholds in order to increase/decrease tundra 
	-- based on tundra level.
	local tundraLatShift = 0.1;
	
	-- Latitude interpolation ranges in [0,1], where 0 == hottest and 1 == uncropped coldest
	local desertMinLat         = sunlightArgs.desertMinLat[climate];
	local desertMaxLat         = sunlightArgs.desertMaxLat[climate];
	local tundraMinLat         = sunlightArgs.tundraMinLat[climate];
	local tundraNonRiverMaxLat = sunlightArgs.tundraNonRiverMaxLat[climate];
	local tundraMaxLat         = sunlightArgs.tundraMaxLat[climate];
	local snowMinLat           = sunlightArgs.snowMinLat[climate];
	local snowMaxLat           = sunlightArgs.snowMaxLat[climate];
	
	-- Offset the plains percentage by user selection of grass-plains mix.
	local plainsMult = DOWN_TO_UP2[grassPlains];
	plainsPct = plainsPct + (plainsMult * 10);
	
	-- Adjust terrain bands according to user's Temperature selection. 
	-- Compute change multiplier based on temperature.
	local tempMult  = DOWN_TO_UP2[temperature];
	
	desertPct = desertPct + (tempMult * desertPctShift);	
	desertMinLat  = desertMinLat  + (tempMult * temperatureLatShift);
	desertMaxLat  = desertMaxLat  + (tempMult * temperatureLatShift);
	tundraMinLat  = tundraMinLat  + (tempMult * temperatureLatShift);
	tundraMaxLat  = tundraMaxLat  + (tempMult * temperatureLatShift);
	snowMinLat    = snowMinLat    + (tempMult * temperatureLatShift);
	snowMaxLat    = snowMaxLat    + (tempMult * temperatureLatShift);
	
	-- Adjust terrain bands that are relevant to tundra level.
	local tundraMult = UP_TO_DOWN2[tundraLevel];
	
	-- Low tundra => wider desert and snow bands, narrower tundra band
	desertMaxLat  = desertMaxLat  + (tundraMult * tundraLatShift);
	tundraMinLat  = tundraMinLat  + (tundraMult * tundraLatShift);
	tundraMaxLat  = tundraMaxLat  + (tundraMult * tundraLatShift);
	snowMinLat    = snowMinLat    - (tundraMult * tundraLatShift);
	
	local data = 
	{
		-- methods
		InitFractals		   = GotLakes_TerrainGenerator.InitFractals,
		InitRegion             = sunlightArgs.TInitRegion,
		GenerateTerrain		   = GotLakes_TerrainGenerator.GenerateTerrain,
		GenerateTerrainAtPlot  = GotLakes_TerrainGenerator.GenerateTerrainAtPlot,
		CreateClimateFractal   = createClimateFracFcn,
		GetLatitudeAtPlot      = latitudeFcn,
		
		-- fields
		iW = iW,
		iH = iH,
		
		desertBottomPct   = 100-desertPct,
		plainsBottomPct   = 100-plainsPct,
		coldPoleBottomPct = 100-coldPolePct,		
		
		desertInterpFcn       = sunlightArgs.desertInterpFcn,
		tundraInterpFcn       = sunlightArgs.tundraInterpFcn,
		snowInterpFcn         = sunlightArgs.snowInterpFcn,
		
		desertMinLat          = desertMinLat,
		desertMaxLat          = desertMaxLat,
		tundraMinLat          = tundraMinLat,
		tundraNonRiverMaxLat  = tundraNonRiverMaxLat,
		tundraMaxLat          = tundraMaxLat,
		snowMinLat	          = snowMinLat,
		snowMaxLat	          = snowMaxLat,
		
		donutFillingMap = donutFillingMap,
		
		gotDonut = (UserInput.landmass == Landmasses.DONUT),
		
		gotDesertDonut = 
			((UserInput.worldWrap == WWRAP_REGION) and 
			 (UserInput.landmass  == Landmasses.DONUT) and 
			 (UserInput.sunlight  == Sunlights.STANDARD) and
			 (UserInput.climate   == Climates.SANDSTORM)),
		 
		gotSnowDonut =
			((UserInput.worldWrap == WWRAP_REGION) and 
			 (UserInput.landmass  == Landmasses.DONUT) and 
			 (UserInput.sunlight  == Sunlights.STANDARD) and
			 (UserInput.climate   == Climates.SNOWBALL)),
			 
		gotGlobe = (UserInput.worldWrap == WWRAP_GLOBE),
	};
	
	data:InitFractals();
	
	return data;
end

function GotLakes_TerrainGenerator:InitFractals()
	-- Initialize terrain fractals and thresholds.	
	self.deserts = self.CreateClimateFractal("deserts");
	self.plains  = self.CreateClimateFractal("plains");
	self.cold    = self.CreateClimateFractal("cold");
	self.polar   = self.CreateClimateFractal("polar");
	
	self.desertTop = GetFractalHeight(self.deserts, 100);
	self.plainsTop = GetFractalHeight(self.plains, 100);
	
	-- Shift some plains away from hills towards flat tiles for playability.
	self.plainsHillBottom = GetFractalHeight(self.plains, self.plainsBottomPct + 25);
	self.plainsFlatBottom = GetFractalHeight(self.plains, self.plainsBottomPct - 8);
end

function GotLakes_TerrainGenerator:Default_InitRegion(enableRandomPoles, defaultR, defaultTheta)
	-- No-op by default. Custom hook for Terrain Generator.
	return {0,0,0,0,0,1};
end

function GotLakes_TerrainGenerator:GenerateTerrain()		
	-- Generate terrain for all non-water plots.
	for x = 0, self.iW - 1 do
		for y = 0, self.iH - 1 do 
			local i = y * self.iW + x;
			local plot = Map.GetPlotByIndex(i);
						
			if (not plot:IsWater()) then
				-- Generate terrain for non-water tiles only.
				-- A previous map-gen step already handled water terrain.
				SetBaseTerrain(plot, self:GenerateTerrainAtPlot(x, y));
			end
		end
	end	
end

function GotLakes_TerrainGenerator:GenerateTerrainAtPlot(iX,iY)
	-- Generate terrain for a non-water tile.
	local lat = self.GetLatitudeAtPlot(iX,iY);
	
	local i = iY * self.iW + iX;
	local plot = Map.GetPlotByIndex(i);
	
	-- Default to a mix of plains and grass (fertile).
	-- Let flat land have a different chance of plains than hills/mtns.
	local plainsBottom = self.plainsFlatBottom;
	if (plot:IsHills() or plot:IsMountain()) then
		plainsBottom = self.plainsHillBottom;
	end	
	local terrainType = g_TERRAIN_TYPE_GRASS;
	local plainsVal = self.plains:GetHeight(iX, iY);
	if ((plainsVal >= plainsBottom) and (plainsVal < self.plainsTop)) then
		terrainType = g_TERRAIN_TYPE_PLAINS;
	end
	
	if (self.gotDesertDonut and (self.donutFillingMap[i+1] == FILLING_YES)) then
		-- Sandstorm donut => desert filling.
		return g_TERRAIN_TYPE_DESERT;
		
	elseif (self.gotSnowDonut and (self.donutFillingMap[i+1] == FILLING_YES)) then
		-- Snowball donut => Snow filling.
		return g_TERRAIN_TYPE_SNOW;
	end
	
	-- Mix of temperate and extremes to improve starts on regional maps.
	local polarVal = self.polar:GetHeight(iX, iY);
	
	-- Chance of desert, depending on latitude.
	local desertBottom = GetFractalHeight(self.deserts, Interpolate(lat, 
		self.desertMinLat, self.desertMaxLat, 
		self.desertBottomPct, 100,
		self.desertInterpFcn));
		
	local desertVal = self.deserts:GetHeight(iX, iY);
	if ((lat <= self.desertMaxLat) and 
		(desertVal >= desertBottom) and 
		(desertVal < self.desertTop)) then
		-- Enforce a hard maximum latitude to avoid fractal anomalies.
		terrainType = g_TERRAIN_TYPE_DESERT;
	end
	
	-- Gradually increase holes in extreme cold as lat increases.
	local coldPoleBottom = GetFractalHeight(self.polar, Interpolate(lat, 
			self.tundraMinLat, self.snowMaxLat, 
			0, self.coldPoleBottomPct,
			self.snowInterpFcn));
	
	if (self.gotGlobe or (polarVal >= coldPoleBottom)) then
		-- Chance of tundra/snow, depending on latitude.
		local snowBottom = GetFractalHeight(self.cold, Interpolate(lat, 
			self.snowMinLat, self.snowMaxLat, 
			100, 0,
			self.snowInterpFcn));
		
		local tundraMaxLat = self.tundraNonRiverMaxLat;
		if (plot:IsRiver()) then
			-- Extend tundra interp for river plots to allow for some temperate
			-- terrain in an otherwise frozen region.
			tundraMaxLat = self.tundraMaxLat;
		end
		
		local tundraBottom = GetFractalHeight(self.cold, Interpolate(lat, 
			self.tundraMinLat, tundraMaxLat, 
			100, 0,
			self.tundraInterpFcn));
		
		local coldVal = self.cold:GetHeight(iX, iY);
		
		if ((lat >= self.snowMinLat) and (coldVal >= snowBottom)) then
			-- Enforce a hard minimum latitude to avoid fractal anomalies.			
			terrainType = g_TERRAIN_TYPE_SNOW;
			
		elseif ((lat >= self.tundraMinLat) and (coldVal >= tundraBottom)) then
			-- Enforce a hard minimum latitude to avoid fractal anomalies.			
			terrainType = g_TERRAIN_TYPE_TUNDRA;
		end
	end
	return terrainType;
end

function GotLakes_TerrainGenerator:TiltedAxis_InitRegion(
	enableRandomPoles, defaultR, defaultTheta)
	-- Tilted Axis regional init.
	-- Args:
	--   enableRandomPoles: whether to randomize poles (true) or use default values (false)
	--   defaultR: default south pole R
	--   defaultTheta: default south pole Theta
	-- Returns:
	--   southPoleX, southPoleY: location of the south pole,
	--	 northPoleX, northPoleY: location of the north pole,
	--	 minR, maxR: min/max distances from south pole (for latitude interpolation)
	print("Initializing Tilted Axis region.");
	
	-- The Euclidian (x,y) location of the center of the map.
	local mapCenterX = math.floor(self.iW/2);
	local mapCenterY = math.floor(self.iH/2);
	if ((mapCenterY % 2) == 1) then
		mapCenterX = mapCenterX + 0.5;
	end
	
	print("Map Center (x,y): " .. CoordString(mapCenterX, mapCenterY));
	
	-- The location of the hot pole that faces the sun, 
	-- relative to the center of the map.
	--
	-- Start with polar coords so that we can randomize by distance and angle.
	
	-- Scale the maximum hot pole distance using whichever dimension is longer.
	local distRange = math.max(self.iW, self.iH) * 5;
	print("Hot Pole Max R: ", distRange);
	
	local hotPoleRDist = defaultR;     -- Hexes
	local hotPoleTheta = defaultTheta; -- Radians
	
	if (self.gotDonut) then
		-- For Donut, fix the hot pole at the center of the map.
		-- This effectively turns the donut filling into desert.
		hotPoleRDist = 0; -- Hexes
		hotPoleTheta = 0; -- Radians
		print("Fixing the hot pole at the donut's filling.");
	
	elseif enableRandomPoles then
		-- In hexes
		hotPoleRDist = TerrainBuilder.GetRandomNumber(distRange,
			"Tilted Axis Region - hot pole distance"); 
		
		-- In Radians
		hotPoleTheta = RandomRadians(0, TAU, 1000);		
	else		
		print("Using default hot pole location (random poles disabled).");
	end
	print("Hot Pole (r, theta): ", CoordString(hotPoleRDist, hotPoleTheta));

	-- The Location of the pole in cartesion coords, so that we can calculate
	-- distance to other (x,y) locations
	local hotPoleXDiff = hotPoleRDist * math.cos(hotPoleTheta);
	local hotPoleYDiff = hotPoleRDist * math.sin(hotPoleTheta);
	
	-- This location may be outside of the hex grid--that is OK!
	local hotPoleX = mapCenterX + hotPoleXDiff;
	local hotPoleY = mapCenterY + hotPoleYDiff;
	print("Hot Pole (x,y): " .. CoordString(hotPoleX, hotPoleY));
	
	-- Find the min/max distances to hot-pole on the map 
	-- so that we can interpolate fractals on a scale that fits with the grid.
	local minR = math.huge;
	local maxR = 0;
	for x = 0, self.iW - 1 do
		for y = 0, self.iH - 1 do 	
			local dist = Distance(x, y, hotPoleX, hotPoleY);
			if (dist > maxR) then
				maxR = dist;
			end
			if (dist < minR) then
				minR = dist;
			end
		end
	end
	print("Hot Pole distance range: " .. CoordString(Round(minR),Round(maxR)));
	
	-- Set global terrain info so that both terrain and feature gen can
	-- share the same hot pole for latitude calculations.
	local data =
	{
		hotPoleX = hotPoleX,
		hotPoleY = hotPoleY,
		minR     = minR,
		maxR     = maxR
	};
	return data;
end

function GotLakes_TerrainGenerator:TwoSuns_InitRegion(enableRandomPoles, defaultR, defaultTheta)
	-- Two Suns regional init.
	-- Args:
	--   enableRandomPoles: whether to randomize poles (true) or use default values (false)
	--   defaultR: default south pole R (ignored)
	--   defaultTheta: default south pole Theta
	-- Returns:
	--   southPoleX, southPoleY: location of the south pole,
	--	 northPoleX, northPoleY: location of the north pole,
	--	 minR, maxR: min/max distances from south pole (for latitude interpolation)
	print("Initializing Two Suns region.");
	
	-- The Euclidian (x,y) location of the center of the map.
	local mapCenterX = math.floor(self.iW/2);
	local mapCenterY = math.floor(self.iH/2);
	if ((mapCenterY % 2) == 1) then
		mapCenterX = mapCenterX + 0.5;
	end
	
	print("Map Center (x,y): " .. CoordString(mapCenterX, mapCenterY));
	
	-- The location of the poles relative to the center of the map.
	-- Start with polar coords so that we can randomize by angle.
	--
	-- Scale the maximum pole distance using whichever dimension is longer.
	local poleRDist = math.max(self.iW, self.iH) * 5;  -- Hexes
	local southPoleTheta = defaultTheta;
	
	if enableRandomPoles then
		southPoleTheta = RandomRadians(0, TAU, 1000); -- Radians
	else
		print("Using default south pole theta (random poles disabled).");
	end
	
	local northPoleTheta = southPoleTheta + math.pi;
	if (northPoleTheta >= TAU) then
		northPoleTheta = northPoleTheta - TAU;
	end
	
	-- The Location of the poles in cartesion coords, so that we can calculate
	-- distance to other (x,y) locations
	local southPoleXDiff = poleRDist * math.cos(southPoleTheta);
	local southPoleYDiff = poleRDist * math.sin(southPoleTheta);
	
	local northPoleXDiff = poleRDist * math.cos(northPoleTheta);
	local northPoleYDiff = poleRDist * math.sin(northPoleTheta);
	
	-- Convert cartesian (x,y) back to a corresponding map-grid theta.
	-- Forces a one-to-one mapping of tiles to thetas for each possible
	-- location of the poles.
	local southPoleTileTheta = math.pi/2 - 
		math.atan2(Round(southPoleXDiff), Round(southPoleYDiff));
	print("South Pole Theta: ", southPoleTileTheta);
	
	-- These locations may be outside of the hex grid--that is OK!
	local southPoleX = mapCenterX + southPoleXDiff;
	local southPoleY = mapCenterY + southPoleYDiff;
	print("South Pole (x,y): " .. CoordString(southPoleX, southPoleY));
	
	local northPoleX = mapCenterX + northPoleXDiff;
	local northPoleY = mapCenterY + northPoleYDiff;
	print("North Pole (x,y): " .. CoordString(northPoleX, northPoleY));
		
	-- Find the min/max distances to nearest pole on the map 
	-- so that we can interpolate fractals on a scale that fits with the grid.
	local minR = math.huge;
	local maxR = 0;
	for x = 0, self.iW - 1 do
		for y = 0, self.iH - 1 do 	
			local dist = math.min(
				Distance(x, y, southPoleX, southPoleY),
				Distance(x, y, northPoleX, northPoleY));
			if (dist > maxR) then
				maxR = dist;
			end
			if (dist < minR) then
				minR = dist;
			end
		end
	end
	print("Hot Pole distance range: " .. CoordString(Round(minR),Round(maxR)));
	
	-- Set global terrain info so that both terrain and feature gen can
	-- share the same hot pole for latitude calculations.
	local data =
	{
		southPoleX = southPoleX,
		southPoleY = southPoleY,
		northPoleX = northPoleX,
		northPoleY = northPoleY,
		minR       = minR,
		maxR       = maxR
	};
	return data;
end